from .celoss import celoss, customceloss
from .celoss import CustomCrossEntropyLoss
from .bceloss import bce, bcelogits, custombce, custombcelogits
from .bceloss import CustomBCELoss, CustomBCEWithLogitsLoss
from .meanloss import mse, mae, custom_mse, custom_mae
from .meanloss import CustomMSELoss, CustomMAELoss
from .base import get_loss
